/*
 * Copyright 2007, Haiku, Inc. All rights reserved.
 * Distributed under the terms of the MIT License.
 *
 * Authors:
 *		Vasilis Kaoutsis, kaoutsis@sch.gr
 */


/*! \brief A simple test of areas creation,
	and deletion, for testing the time-costs.
*/


#include <Application.h>
#include <OS.h>
#include <String.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/utsname.h>
#include <sys/time.h>
#include <sys/times.h>
#include <unistd.h>

//#define TRACE_ALL
#ifdef TRACE_ALL
#	define TRACE_AREA_CREAT_DELETE
#	define TRACE_NULL_FUNCTION
#	define TRACE_ITERATIONS
#	define TRACE_PAGE_TOUCHING
#endif

#ifdef TRACE_AREA_CREAT_DELETE
#	define TR_AR_CR_DL(x) fprintf x
#else
#	define TR_AR_CR_DL(x) ;
#endif

#ifdef TRACE_NULL_FUNCTION
#	define TR_NL_FC(x) fprintf x
#else
#	define TR_NL_FC(x) ;
#endif

#ifdef TRACE_ITERATIONS
#	define TR_ITER(x) fprintf x
#else
#	define TR_ITER(x) ;
#endif

#ifdef TRACE_PAGE_TOUCHING
#	define TR_TOUCH(x) fprintf x
#else
#	define TR_TOUCH(x) ;
#endif

void DoNothing()
{
}
	
enum measuring_t {
	kUsingGettimeofday = 1,
	kUsingSystemTime,
	kAll
};

const char* kSignature = "application/x-vnd.Haiku.cmd-area_creation_test";
const char* kProgramName = "area_creation_test";

const int32 kErrorInitFail = 127;
const int32 kErrorArgumentsFail = 126;


class AreaTestApplication;
typedef status_t (AreaTestApplication::*PMFI)(uint32 iterations);


class AreaTestApplication : public BApplication {
	public:
		AreaTestApplication();
		virtual ~AreaTestApplication() { }

		virtual void ReadyToRun();
		virtual void ArgvReceived(int32 argc, char** argv);

		bool GoodArguments() const { return fOk; }
		status_t AreaCreate(uint32 numOfAreas);
		status_t DeleteAllAreas(uint32 numOfAreas);
		status_t DeleteEvenAreas(uint32 numOfAreas);
		status_t DeleteOddAreas(uint32 numOfAreas);
		status_t TouchPages(uint32 numOfAreas);
		status_t NullFunction(uint32 counter);
		unsigned long UsingGettimeofday(PMFI);
		bigtime_t UsingSystemTime(PMFI);
		void PrintResults(measuring_t measuringType, PMFI);

	private:
		void _Usage() const;

	private:
		bool fOk;
		void* fStartingAddress;
		uint32 fAddressSpec;
		uint32 fRequestedSize;
		uint32 fLockSpec;
		uint32 fProtection;
		uint32 fNumOfPages;
		BString fArchitecture;
		BString fFileName;
		uint32 fResultSum;
		uint32 fResult;
		uint32 fIterations;
		uint32 fNumOfAreas;
		uint32 fAreaIdTag[5000];
		uint32 fAreaPrepIdTag[5000];
		uint8* fAreaAddress[5000];
		bool fFragmentationTest;

};


AreaTestApplication::AreaTestApplication()
	: BApplication(kSignature),
	fOk(false),
	fStartingAddress(NULL),
	fAddressSpec(0),
	fRequestedSize(0),
	fLockSpec(0),
	fProtection(0),
	fNumOfPages(0),
	fResultSum(0),
	fResult(0),
	fIterations(fNumOfAreas),
	fFragmentationTest(false)
{
}


void
AreaTestApplication::_Usage() const
{
	fprintf(stdout,"\nUsage:\n"
		"%s [pages number] [address spec] [lock spec] [protection] \n\n"
		" [number of areas to create] [fragm] \n\n"
		"\t [pages number] is an integer value.\n"
		"\tBy providing the number of pages, you provide the size of the area.\n"
		"\tThe size of the area is calculating as: B_PAGE_SIZE * numOfPages\n"
		"\tB_PAGE_SIZE has the value of one page or 4096 bytes.\n"
		"\tFor example if you want the size of the area to be 40 kB.\n"
		"\tyour argument have to be 10, since 4096 * 10 is 40960 bytes or 40 kB.\n"
		"\n"
		"\t [addreess spec] is one of the following values:\n"
		"\tB_EXACT_ADDRESS, B_BASE_ADDRESS, B_ANY_ADDRESS, B_ANY_KERNEL_ADDRESS\n\n"
		"\t [lock spec] is one of the following values:\n"
		"\tB_NO_LOCK, B_LAZY_LOCK, B_FULL_LOCK, B_CONTIGUOUS, B_LOMEM\n\n"
		"\t [protection] is one of the following values:\n"
		"\tB_READ_AREA, B_WRITE_AREA, B_READ_AREA | B_WRITE_AREA\n\n"
		"\t [number of areas to create] max value is currently 5000 \n\n"
		"\t fragm: run the fragmentation test, optional\n"
		"An example:\n"
		"area_creation_test.beos 10 B_ANY_ADDRESS B_NO_LOCK 'B_WRITE_AREA|B_READ_AREA' 4000\n"
		"will create 4000 areas that each one will have 10 pages size.\n"
		"Another example:\n"
		"area_creation_test.beos 10 B_ANY_ADDRESS B_NO_LOCK 'B_WRITE_AREA|B_READ_AREA' 4000 fragm\n"
		"will create 4000 areas that each one will have 10 pages size,\n"
		"plus the fragmentation test\n"
		, kProgramName);
}


void
AreaTestApplication::PrintResults(measuring_t measuringType, PMFI BenchFunction)
{
	switch (measuringType) {
		static bool usingAll;

		case kAll:
			usingAll = true;

		case kUsingGettimeofday:
			fprintf(stdout, "gettimeofday");
			fResultSum = UsingGettimeofday(BenchFunction);

			fResult = fResultSum / fIterations;
			if (fResult == 0) { // result is ns
				fResult = 1000 * fResultSum / fIterations;
				fprintf(stdout, "%11ld ns", fResult);
			} else
				fprintf(stdout, "%11ld µs.", fResult);

			if (usingAll == false)
				break;
			fprintf(stdout, "\n");

		case kUsingSystemTime:
			fprintf(stdout, "system_time");
			fResultSum = UsingSystemTime(BenchFunction);

			fResult = fResultSum / fIterations;
			if (fResult == 0) { // result is ns
				fResult = (1000 * fResultSum) / fIterations;
				fprintf(stdout, "%12ld ns", fResult);
			} else
				fprintf(stdout, "%12ld µs.", fResult);

			if (usingAll == false)
				break;
			fprintf(stdout, "\n");

	}
	fprintf(stdout, "\n");
}


void
AreaTestApplication::ArgvReceived(int32 argc, char** argv)
{
	if (argc < 2 || argc > 7)
		return; // bad arguments number!

	fNumOfPages = atoi(argv[1]);
	
	if (argc < 6)
		fNumOfAreas = 5000;
	else
		fNumOfAreas = atoi(argv[5]);

	if (fNumOfAreas > 5000) {
		fprintf(stdout, "Currently the test doesn't support the creation of more than 5000 areas!");
		return;
	}

	fIterations = fNumOfAreas;

	if (argc == 7 && argv[6][0] == 'f'
		&& argv[6][1] == 'r'
		&& argv[6][2] == 'a'
		&& argv[6][3] == 'g'
		&& argv[6][4] == 'm')
		fFragmentationTest = true;

	// get the addressSpec
	BString addressSpecString(argv[2]);
	if (addressSpecString == "B_EXACT_ADDRESS")
		fAddressSpec = 1;
	else if (addressSpecString == "B_BASE_ADDRESS")
		fAddressSpec = 2;
	else if (addressSpecString == "B_ANY_ADDRESS")
		fAddressSpec = 0;
	else if (addressSpecString == "B_ANY_KERNEL_ADDRESS")
		fAddressSpec = 4;
	else {
		fprintf(stdout, "Wrong address specification: %s\n"
			"\taddreess specification is one of the following values:\n"
			"\tB_EXACT_ADDRESS, B_BASE_ADDRESS, B_ANY_ADDRESS, B_ANY_KERNEL_ADDRESS\n\n", argv[2]);
		return;
	}

	// get the lock specification
	BString lockSpecString(argv[3]);
	if (lockSpecString == "B_NO_LOCK")
		fLockSpec = 0;
	else if (lockSpecString == "B_LAZY_LOCK")
		fLockSpec = 1;
	else if (lockSpecString == "B_FULL_LOCK")
		fLockSpec = 2;
	else if (lockSpecString == "B_CONTIGUOUS")
		fLockSpec = 3;
	else if (lockSpecString == "B_LOMEM")
		fLockSpec = 4;
	else {
		fprintf(stdout, "Wrong lock specification: %s\n"
			"\tlock specification is one of the following values:\n"
			"\tB_NO_LOCK, B_LAZY_LOCK, B_FULL_LOCK, B_CONTIGUOUS, B_LOMEM\n\n", argv[3]);
		return;
	}
	
	// get the protection flags
	BString protectionFlagStr(argv[4]);
	if (protectionFlagStr == "B_READ_AREA") {
		fprintf(stdout,
			"WARNING: area is read only.\n"
			"Any attempt for writing should be fail!\n"
			"Expecting crash!\n");
		fProtection = 1;
	} else if (protectionFlagStr == "B_WRITE_AREA")
		fProtection = 2;
	else if (protectionFlagStr == "B_READ_AREA | B_WRITE_AREA"
		|| protectionFlagStr == "B_READ_AREA|B_WRITE_AREA"
		|| protectionFlagStr == "B_READ_AREA |B_WRITE_AREA"
		|| protectionFlagStr == "B_READ_AREA| B_WRITE_AREA"
		|| protectionFlagStr == "B_WRITE_AREA | B_READ_AREA"
		|| protectionFlagStr == "B_WRITE_AREA|B_READ_AREA"
		|| protectionFlagStr == "B_WRITE_AREA |B_READ_AREA"
		|| protectionFlagStr == "B_WRITE_AREA| B_READ_AREA")
		fProtection = 3;
	else {
		fprintf(stdout, "Wrong protection flags: %s\n"
			"\t protection is one of the following values:\n"
			"\tB_READ_AREA, B_WRITE_AREA, B_READ_AREA | B_WRITE_AREA\n", argv[4]);
		return;
	}

	fOk = true;
}


status_t
AreaTestApplication::AreaCreate(uint32 numOfAreas)
{
	TR_AR_CR_DL((stdout, "numOfAreas is %ld\n", numOfAreas));
	char error;
	if (numOfAreas == 0) {
		error = B_ERROR;
		throw error;
	}

	static bool fragmPreparation = true;

	for (uint32 indx = 0 ; indx < numOfAreas; indx++) {
		area_id areaSpace;
		areaSpace = create_area("AreaSpace", &fStartingAddress, fAddressSpec,
			fRequestedSize, fLockSpec, fProtection);

		if (fFragmentationTest && fragmPreparation) {
			fAreaPrepIdTag[indx] = fAreaIdTag[indx] = areaSpace;
				// fAreaPrepIdTag is an array that will be used
				// for deleting the remaining preparation areas
		} else
			fAreaIdTag[indx] = areaSpace;

		TR_AR_CR_DL((stdout, "fAreaIdTag[%ld] is %ld at %p\n", indx, fAreaIdTag[indx],
			fStartingAddress));

		if (areaSpace < 0) {
			fprintf(stdout, "Fail to create area!\n");
			fprintf(stdout, "fAreaIdTag[%ld] is %ld at %p\n", indx, fAreaIdTag[indx],
				fStartingAddress);
			fprintf(stdout, "areaSpace is %ld\n", areaSpace);
			throw (area_id)(areaSpace);
		}

		fAreaAddress[indx] = static_cast<uint8*>(fStartingAddress);
		fStartingAddress = static_cast<uint8*>(fStartingAddress) + fRequestedSize;
	}

	fStartingAddress = static_cast<uint8*>(fAreaAddress[0]);

	if (fFragmentationTest)
		fragmPreparation = false;

	return B_OK;
}


status_t
AreaTestApplication::TouchPages(uint32 numOfAreas)
{
	// Touching requested page(s) of each area.

	// although it might seems not exact
	// this function measures only the writing to pages (page fault handling)
	// namely this: *areaPointer = 1;
	// comment out the writing, this function
	// for 500 areas (100 pages per area) costs 310 nanosec per area!

	for (uint32 index = 0; index < numOfAreas; index++) {
		uint8* areaPointer = fAreaAddress[index];
		for (uint32 indx = 0 ; indx < fNumOfPages; indx++) {
			*areaPointer = 1;
			TR_TOUCH((stdout, "fAreaAddress[%d] is %p, fNumOfPages are %d,"
				" writing at %p, the value of %d\n", index, fAreaAddress[index], fNumOfPages,
				areaPointer, *areaPointer));
			areaPointer = areaPointer + B_PAGE_SIZE;
		}
	}
	return B_OK;
}


status_t
AreaTestApplication::DeleteAllAreas(uint32 numOfAreas)
{
	status_t error;
	for (uint32 index = 0 ; index < numOfAreas; index++) {
		TR_AR_CR_DL((stdout, "deleting area with fAreaIdTag[%ld] %ld \n", index, fAreaIdTag[index]));
		error = delete_area(fAreaIdTag[index]);
		if (error == B_ERROR) {
			fprintf(stdout, "Couldn't delete area with fAreaIdTag[%ld] %ld \n", index, fAreaIdTag[index]);
			throw (status_t)(error);
		}	
	}
	return B_OK;
}


status_t
AreaTestApplication::DeleteEvenAreas(uint32 numOfAreas)
{
	status_t error;
	for (uint32 index = 0 ; index < numOfAreas; index += 2) {
		TR_AR_CR_DL((stdout, "deleting area with fAreaPrepIdTag[%ld] %ld \n", index, fAreaPrepIdTag[index]));
		error = delete_area(fAreaPrepIdTag[index]);
		if (error == B_ERROR) {
			fprintf(stdout, "Couldn't delete area with fAreaPrepIdTag[%ld] %ld \n", index, fAreaPrepIdTag[index]);
			throw (status_t)(error);
		}	
	}
	return B_OK;
}


status_t
AreaTestApplication::DeleteOddAreas(uint32 numOfAreas)
{
	status_t error;
	for (uint32 index = 1 ; index < numOfAreas; index += 2) {
		TR_AR_CR_DL((stdout, "deleting area with fAreaPrepIdTag[%ld] %ld \n", index, fAreaPrepIdTag[index]));
		error = delete_area(fAreaPrepIdTag[index]);
		if (error == B_ERROR) {
			fprintf(stdout, "Couldn't delete area with fAreaPrepIdTag[%ld] %ld \n", index, fAreaPrepIdTag[index]);
			throw (status_t)(error);
		}	
	}
	return B_OK;
}

status_t
AreaTestApplication::NullFunction(uint32 counter)
{
	fIterations = 5000000;
	TR_NL_FC((stdout, "inside NullFunction, counter is %ld.\n", counter));
	// the only usefulness of this function
	// is to measure function call overhead
	for (uint32 index = 0; index < 5000000; index++)
		DoNothing();
	
	return B_OK;
}


unsigned long
AreaTestApplication::UsingGettimeofday(PMFI BenchFunction)
{
	struct timeval before;
	struct timeval after;
	unsigned long elapsed;
	
	gettimeofday(&before, NULL);
	(this->*BenchFunction)(fIterations);
	gettimeofday(&after, NULL);

	
	elapsed = 1000000 * (after.tv_sec - before.tv_sec);
	elapsed += after.tv_usec - before.tv_usec;
	TR_ITER((stdout, "fIterations is %d", fIterations));
	return elapsed;
}


bigtime_t
AreaTestApplication::UsingSystemTime(PMFI BenchFunction)
{
	bigtime_t start;
	bigtime_t stop;

	start = system_time();
	(this->*BenchFunction)(fIterations);
	stop = system_time();
	
	TR_ITER((stdout, "fIterations is %d", fIterations));
	return stop - start;
}


void
AreaTestApplication::ReadyToRun()
{
	if (GoodArguments()) {

		// create the area	
		area_id areaSpace;
		fRequestedSize = B_PAGE_SIZE * fNumOfPages; // B_PAGE_SIZE = 4096 bytes
		void* startingAddress = (void*)fRequestedSize;
		areaSpace = create_area("AreaSpace", &startingAddress, fAddressSpec,
			fRequestedSize, fLockSpec, fProtection);


		// errors checking

		if (areaSpace < 0)
			fprintf(stdout, "The system said: %s. Error code number is %ld\n", strerror(areaSpace), areaSpace);

		if (areaSpace == B_BAD_VALUE) {
			// SUGESTION: here we may want in the future to extend our create_area
			// implemetation for the bad values,
			// to see what excactly was going on.
			fprintf(stdout, "Bad argument value.\n"
				"Possible errors:\n"
				" a) You passed an unrecognized constant for addr_spec or lock\n"
				" b) the addr or size value isn't a multiple of B_PAGE_SIZE,\n"
				" c) you set addr_spec to B_EXACT_ADDRESS but the address request couldn't\n"
				" be fulfilled. \n");
			Quit();
			return;
		}

		if (areaSpace == B_NO_MEMORY)  {
			if (fRequestedSize < (900 * 4096))
				fprintf(stdout, "Requested size %ld KB is to big!\n", fRequestedSize / 1024);
			else
				fprintf(stdout, "Requested size %ld MB is to big!\n", fRequestedSize / (1024 * 1024));
			Quit();
			return;
		}

		if (areaSpace == B_ERROR) {
			fprintf(stdout, "A system error prevented the area from being created.\n");
			Quit();
			return;
		}


		// area has pass errors checking, start the tests

		delete_area(areaSpace);

		// show sysname and release
		struct utsname osName;
		status_t error = uname(&osName);
		if (error != B_ERROR)
			fprintf(stdout, "\nSysname: %s\trelease: %s\tversion: %s\n",
				osName.sysname, osName.release, osName.version);
		else {
			fprintf(stdout, "\nUnable to get system name and release!\n");
			fprintf(stdout, "Error was %s, and error number is %d\n", strerror(error), error);
		}

		// show wheather we are running the fragmentation
		// test or not
		if (fFragmentationTest) {
			fprintf(stdout, "Running the fragmentation test: \n"
				"Creating preparation areas; will be deleted at even indices.\n"
				"(The remaining areas will be deleted after the measuring tests!)\n");
		}

		// show user's request
		if (fRequestedSize < (1024 * 1024))
			fprintf(stdout,
				"\nRequested %ld areas. Area's size will be %.2f KB (%d pages).\n",
				fNumOfAreas, fRequestedSize / double(1024), fNumOfPages);		
		else
			fprintf(stdout,
				"\nRequested %ld areas. Area's size will be %.2f MB (%d pages).\n",
				fNumOfAreas, fRequestedSize / double(1024 * 1024), fNumOfPages);

		// open the file *.data
		// for writing some infos
		FILE* pointerToDataFile;
		BString dataFileName(fFileName.String());
		dataFileName += ".data";
		pointerToDataFile = fopen(dataFileName.String(), "w");

		try {
			AreaCreate(fNumOfAreas);
 
 			char* showAddress = static_cast<char*>(fStartingAddress);
 			for (uint32 index = 0; index < fNumOfAreas; index++) {
				fprintf(pointerToDataFile, "fAreaIdTag[%ld] area_id is %ld starting at %p\n",
					index, fAreaIdTag[index], showAddress);
				showAddress = showAddress + fRequestedSize;
			}

		} catch(char& error) {
			fprintf(stdout, "abording time tests, reason: number of areas is 0!\n");
			fclose(pointerToDataFile);
			Quit();
			return;
		} catch(area_id& areaIdValue) {
			fprintf(stdout, "abording time tests!\n");
			fprintf(stdout, "areaIdValue is %ld \n", areaIdValue);
			fclose(pointerToDataFile);
			DeleteAllAreas(fNumOfAreas);
			Quit();
			return;
		}

		if (fFragmentationTest) {
			try {
				DeleteEvenAreas(fNumOfAreas);
			} catch(status_t& deleteAreaError) {
				fprintf(stdout, "Abording time tests!\n");
				fprintf(stdout, "deleteAreaError is %ld\n", deleteAreaError);
			}
		} else {
			try {
				DeleteAllAreas(fNumOfAreas);
			} catch(status_t& deleteAreaError) {
				fprintf(stdout, "Abording time tests!\n");
				fprintf(stdout, "deleteAreaError is %ld\n", deleteAreaError);
			}
		}

		fclose(pointerToDataFile);

		fprintf(stdout, "\n");
		fprintf(stdout, "------------------------------------------------------------------\n");
		fprintf(stdout, "        Measuring the time of area's creation and deletion\n");
		fprintf(stdout, "------------------------------------------------------------------\n");

		// Print results
		try {
			fprintf(stdout, "%1s %15s", "TimeFunction", "AreaCreate\n");
			PrintResults(kUsingGettimeofday, (&AreaTestApplication::AreaCreate));
			fprintf(stdout, "%1s %15s", "TimeFunction", "TouchPages\n");
			PrintResults(kUsingGettimeofday, (&AreaTestApplication::TouchPages));
			fprintf(stdout, "%1s %15s", "TimeFunction", "DeleteAllAreas\n");
			PrintResults(kUsingGettimeofday, (&AreaTestApplication::DeleteAllAreas));
			fprintf(stdout, "------------------------------------\n");
		} catch(area_id& areaIdVal) {
			fprintf(stdout, "abording time tests!\n");
			fprintf(stdout, "areaIdVal is %ld \n", areaIdVal);
		} catch(status_t& deleteAreaError) {
			fprintf(stdout, "Abording time tests!\n");
			fprintf(stdout, "deleteAreaError is %ld\n", deleteAreaError);
		}

		// the following is only to keep an eye to the NuullFunctions timings
		//PrintResults(kAll, (&AreaTestApplication::NullFunction));

		if (fFragmentationTest) {
			try {
				DeleteOddAreas(fNumOfAreas);
			} catch(status_t& deleteAreaError) {
				fprintf(stdout, "Abording time tests!\n");
				fprintf(stdout, "deleteAreaError is %ld\n", deleteAreaError);
			}
		}

	} else
		_Usage();

	Quit();
}


//	#pragma mark -


int
main(int32 argc, const char** argv)
{
	AreaTestApplication app;
	if (app.InitCheck() != B_OK)
		return kErrorInitFail;

	app.Run();
	if (!app.GoodArguments())
		return kErrorArgumentsFail;

	return 0;
}
